# -*- coding: utf-8 -*-
from celery.task import Task
from vkontakte_groups_migration.models import GroupMigration

class VkontakteGroupUpdateUsersM2M(Task):

    def run(self, stat_id, **kwargs):
        stat = GroupMigration.objects.get(pk=stat_id)
        logger = self.get_logger(**kwargs)
        logger.info(u'VK group "%s" users m2m relations updating started' % stat.group)
        try:
            GroupMigration.objects.update_group_users_m2m(stat)
            logger.info(u'VK group "%s" users m2m relations succesfully updated' % stat.group)
        except:
            logger.error(u'Unknown error while updating users m2m relations of VK group "%s"' % stat.group)
        return True


# -*- coding: utf-8 -*-
from django.db import models
from django.db.models.query import QuerySet
from django.core.exceptions import MultipleObjectsReturned
from twitter_api import fields
from twitter_api.utils import api
from twitter_api.models import User
from datetime import datetime
import logging

log = logging.getLogger('twitter_relations_history')

class NoDeltaForFirstHistory(ValueError):
    pass

def ModelQuerySetManager(ManagerBase=models.Manager):
    '''
    Function that return Manager for using QuerySet class inside the model definition
    @param ManagerBase - parent class Manager
    '''
    if not issubclass(ManagerBase, models.Manager):
        raise ValueError("Parent class for ModelQuerySetManager must be models.Manager or it's child")

    class Manager(ManagerBase):
        '''
        Manager based on QuerySet class inside the model definition
        '''
        def get_query_set(self):
            return self.model.QuerySet(self.model)

    return Manager()

class RelationsHistoryQueryset(object):

    @property
    def visible(self):
        return self.exclude(hidden=True).exclude(time__isnull=True)

    @property
    def light(self):
        return self.defer(
            'followers_ids',
        )

class RelationsHistoryManager(models.Manager, RelationsHistoryQueryset):

    def update_for_user(self, user, offset=0):
        '''
        Fetch all users for this user, save them as IDs and after make m2m relations
        '''
        try:
            stat, created = self.get_or_create(user=user, time=None)
        except MultipleObjectsReturned:
            self.filter(user=user, time=None).delete()
            stat = self.create(user=user, time=None, offset=0)
            created = True

        if created:
            stat.set_defaults()

        stat.time = datetime.now()
        stat.followers_ids = user.fetch_followers_ids(all=True)
        stat.update()
        stat.save()

#        signals.group_migration_updated.send(sender=RelationsHistory, instance=stat)

#     def update_group_users_m2m(self, stat, offset=0):
#         '''
#         Fetch all users of group, make new m2m relations, remove old m2m relations
#         '''
#         group = stat.group
#
#         ids = stat.followers_ids
#         ids_left = set(group.users.values_list('remote_id', flat=True)).difference(set(ids))
#
#         offset = offset or stat.offset
#         errors = 0
#
#         while True:
#             ids_sliced = ids[offset:offset+1000]
#             if len(ids_sliced) == 0:
#                 break
#
#             log.debug('Fetching users for group "%s", offset %d' % (group, offset))
#             try:
#                 users = User.remote.fetch(ids=ids_sliced, only_expired=True)
#             except Exception, e:
#                 log.error('Error %s while getting users for group "%s": "%s", offset %d' % (e.__class__, group, e, offset))
#                 errors += 1
#                 if errors == 10:
#                     log.error('Fail - number of unhandled errors while updating users for group "%s" more than 10, offset %d' % (group, offset))
#                     break
#                 continue
#
#             if len(users) == 0:
#                 stat.offset = 0
#                 stat.save()
#                 break
#             else:
#                 for user in users:
#                     if user.id:
#                         group.users.add(user)
#                 stat.offset = offset
#                 stat.save()
#                 offset += 1000
#
#         # process left users of group
#         log.info('Removing %s left users for group "%s"' % (len(ids_left), group))
#         for remote_id in ids_left:
#             group.users.remove(User.objects.get(remote_id=remote_id))
#
#         signals.group_users_updated.send(sender=Group, instance=group)
#         log.info('Updating m2m relations of users for group "%s" successfuly finished' % (group,))
#         return True

class RelationsHistory(models.Model):
    class Meta:
        verbose_name = u'Relations history of twitter users'
        verbose_name_plural = u'Relations histories of twitter users'
        unique_together = ('user','time')
        ordering = ('user','time','-id')

    class QuerySet(QuerySet, RelationsHistoryQueryset):
        pass

#     class MethodDelta(object):
#         def __init__(self, client, method_name):
#             self.client = client
#             self.method_name = method_name
#
#         def __getattr__(self, key):
#             return RelationsHistory.MethodDelta(self.client, '.'.join((self.method_name, key)))
#
#         def __call__(self, prev=None):
#             if not prev:
#                 prev = self.prev
#             if prev:
#                 delta = RelationsHistoryDelta.objects.get_or_create(history_from=prev, history_to=self)[0]
#                 return getattr(delta, self.method_name)
#             else:
#                 raise NoDeltaForFirstHistory("This history doesn't have previous history")

    user = models.ForeignKey(User, verbose_name=u'User', related_name='relations_history')
    time = models.DateTimeField(u'Datetime', null=True)

    hidden = models.BooleanField(u'Скрыть')

    offset = models.PositiveIntegerField(default=0)

    followers_ids = fields.PickledObjectField(default=[])
    followers_count = models.PositiveIntegerField(default=0)

    objects = ModelQuerySetManager(RelationsHistoryManager)

    def set_defaults(self):
        '''
        It's neccesary to call after creating of every instance,
        because `default` attribute of fields.PickledObjectField doesn't work properly
        '''
        self.followers_ids = []

    _next = None
    _prev = None

    @property
    def next(self):
        try:
            assert self._next
        except AssertionError:
            try:
                self._next = self.user.relations_history.visible.filter(time__gt=self.time).order_by('time')[0]
            except IndexError:
                return None
        return self._next

    @property
    def prev(self):
        try:
            assert self._prev
        except AssertionError:
            try:
                self._prev = self.user.relations_history.visible.filter(time__lt=self.time).order_by('-time')[0]
            except IndexError:
                return None
        return self._prev

    def get_delta(self, name, prev=None):
        if name in [field.name for field in RelationsHistoryDelta._meta.fields if 'followers' in field.name]:
            if not prev:
                prev = self.prev
            if prev:
                delta = RelationsHistoryDelta.objects.get_or_create(history_from=prev, history_to=self)[0]
                return getattr(delta, name)
            else:
                raise NoDeltaForFirstHistory("This history doesn't have previous history")
        else:
            raise AttributeError("type object '%s' has no attribute '%s'" % (self.__class__, name))

    def __getattr__(self, name):
        return self.get_delta(name)
#        return RelationsHistory.MethodDelta(self, name)

    def delete(self, *args, **kwargs):
        '''
        Recalculate next stat members instance
        '''
        self.hide()
        super(RelationsHistory, self).delete(*args, **kwargs)

    def hide(self):
        '''
        Hide curent migration, and recalculate fields of next migrations
        '''
        self.hidden = True
        self.save()

    def update_next(self):
        next_stat = self.next
        if next_stat:
            next_stat.update()
            next_stat.save()

    def save(self, *args, **kwargs):
        update_next = False
        if self.id and self.hidden != self.__class__.objects.light.get(id=self.id).hidden:
            update_next = True

        super(RelationsHistory, self).save(*args, **kwargs)

        if update_next:
            self.update_next()

    def update(self):
        self.update_migration()
        self.update_counters()

    def update_migration(self):
        return
        prev_stat = self.prev
        if prev_stat and self.user:
            self.members_left_ids = list(set(prev_stat.followers_ids).difference(set(self.followers_ids)))
            self.members_entered_ids = list(set(self.followers_ids).difference(set(prev_stat.followers_ids)))

    def update_counters(self):
        for field_name in ['followers']:
            setattr(self, field_name + '_count', len(getattr(self, field_name + '_ids')))

class RelationsHistoryDelta(models.Model):
    class Meta:
        unique_together = ('history_from','history_to')

    history_from = models.ForeignKey(RelationsHistory, related_name='deltas_from')
    history_to = models.ForeignKey(RelationsHistory, related_name='deltas_to')

    followers_left_ids = fields.PickledObjectField(default=[])
    followers_entered_ids = fields.PickledObjectField(default=[])

    followers_left_count = models.PositiveIntegerField(default=0)
    followers_entered_count = models.PositiveIntegerField(default=0)

    def save(self, *args, **kwargs):

        self.followers_left_ids = list(set(self.history_from.followers_ids).difference(set(self.history_to.followers_ids)))
        self.followers_entered_ids = list(set(self.history_to.followers_ids).difference(set(self.history_from.followers_ids)))

        self.followers_left_count = len(self.followers_left_ids)
        self.followers_entered_count = len(self.followers_entered_ids)

        super(RelationsHistoryDelta, self).save(*args, **kwargs)

import signals


# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'RelationsHistory'
        db.create_table('twitter_relations_history_relationshistory', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(related_name='relations_history', to=orm['twitter_api.User'])),
            ('time', self.gf('django.db.models.fields.DateTimeField')(null=True)),
            ('hidden', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('offset', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('followers_ids', self.gf('picklefield.fields.PickledObjectField')(default=[])),
            ('followers_count', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
        ))
        db.send_create_signal('twitter_relations_history', ['RelationsHistory'])

        # Adding unique constraint on 'RelationsHistory', fields ['user', 'time']
        db.create_unique('twitter_relations_history_relationshistory', ['user_id', 'time'])

        # Adding model 'RelationsHistoryDelta'
        db.create_table('twitter_relations_history_relationshistorydelta', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('history_from', self.gf('django.db.models.fields.related.ForeignKey')(related_name='deltas_from', to=orm['twitter_relations_history.RelationsHistory'])),
            ('history_to', self.gf('django.db.models.fields.related.ForeignKey')(related_name='deltas_to', to=orm['twitter_relations_history.RelationsHistory'])),
            ('followers_left_ids', self.gf('picklefield.fields.PickledObjectField')(default=[])),
            ('followers_entered_ids', self.gf('picklefield.fields.PickledObjectField')(default=[])),
            ('followers_left_count', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
            ('followers_entered_count', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
        ))
        db.send_create_signal('twitter_relations_history', ['RelationsHistoryDelta'])

        # Adding unique constraint on 'RelationsHistoryDelta', fields ['history_from', 'history_to']
        db.create_unique('twitter_relations_history_relationshistorydelta', ['history_from_id', 'history_to_id'])


    def backwards(self, orm):
        # Removing unique constraint on 'RelationsHistoryDelta', fields ['history_from', 'history_to']
        db.delete_unique('twitter_relations_history_relationshistorydelta', ['history_from_id', 'history_to_id'])

        # Removing unique constraint on 'RelationsHistory', fields ['user', 'time']
        db.delete_unique('twitter_relations_history_relationshistory', ['user_id', 'time'])

        # Deleting model 'RelationsHistory'
        db.delete_table('twitter_relations_history_relationshistory')

        # Deleting model 'RelationsHistoryDelta'
        db.delete_table('twitter_relations_history_relationshistorydelta')


    models = {
        'twitter_api.user': {
            'Meta': {'object_name': 'User'},
            'contributors_enabled': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {}),
            'default_profile': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'default_profile_image': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'description': ('django.db.models.fields.TextField', [], {}),
            'entities': ('annoying.fields.JSONField', [], {}),
            'favorites_count': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'fetched': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'follow_request_sent': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'followers': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'followings'", 'symmetrical': 'False', 'to': "orm['twitter_api.User']"}),
            'followers_count': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'following': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'friends_count': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'geo_enabled': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'id': ('django.db.models.fields.BigIntegerField', [], {'primary_key': 'True'}),
            'is_translator': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'lang': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            'listed_count': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'location': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'notifications': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'profile_background_color': ('django.db.models.fields.CharField', [], {'max_length': '6'}),
            'profile_background_image_url': ('django.db.models.fields.URLField', [], {'max_length': '300'}),
            'profile_background_image_url_https': ('django.db.models.fields.URLField', [], {'max_length': '300'}),
            'profile_background_tile': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'profile_banner_url': ('django.db.models.fields.URLField', [], {'max_length': '300'}),
            'profile_image_url': ('django.db.models.fields.URLField', [], {'max_length': '300'}),
            'profile_image_url_https': ('django.db.models.fields.URLField', [], {'max_length': '300'}),
            'profile_link_color': ('django.db.models.fields.CharField', [], {'max_length': '6'}),
            'profile_sidebar_border_color': ('django.db.models.fields.CharField', [], {'max_length': '6'}),
            'profile_sidebar_fill_color': ('django.db.models.fields.CharField', [], {'max_length': '6'}),
            'profile_text_color': ('django.db.models.fields.CharField', [], {'max_length': '6'}),
            'profile_use_background_image': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'protected': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'screen_name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '50'}),
            'statuses_count': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'time_zone': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True'}),
            'url': ('django.db.models.fields.URLField', [], {'max_length': '300', 'null': 'True'}),
            'utc_offset': ('django.db.models.fields.IntegerField', [], {'null': 'True'}),
            'verified': ('django.db.models.fields.BooleanField', [], {'default': 'False'})
        },
        'twitter_relations_history.relationshistory': {
            'Meta': {'ordering': "('user', 'time', '-id')", 'unique_together': "(('user', 'time'),)", 'object_name': 'RelationsHistory'},
            'followers_count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'followers_ids': ('picklefield.fields.PickledObjectField', [], {'default': '[]'}),
            'hidden': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'offset': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'time': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'relations_history'", 'to': "orm['twitter_api.User']"})
        },
        'twitter_relations_history.relationshistorydelta': {
            'Meta': {'unique_together': "(('history_from', 'history_to'),)", 'object_name': 'RelationsHistoryDelta'},
            'followers_entered_count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'followers_entered_ids': ('picklefield.fields.PickledObjectField', [], {'default': '[]'}),
            'followers_left_count': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'followers_left_ids': ('picklefield.fields.PickledObjectField', [], {'default': '[]'}),
            'history_from': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'deltas_from'", 'to': "orm['twitter_relations_history.RelationsHistory']"}),
            'history_to': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'deltas_to'", 'to': "orm['twitter_relations_history.RelationsHistory']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'})
        }
    }

    complete_apps = ['twitter_relations_history']


# -*- coding: utf-8 -*-
from django.contrib import admin
from vkontakte_api.admin import VkontakteModelAdmin
from vkontakte_groups.admin import Group, GroupAdmin as GroupAdminOriginal
from models import GroupMigration

class GroupMigrationInline(admin.TabularInline):
    model = GroupMigration
    fields = ('group','time','offset','hidden','members_count','members_entered_count','members_left_count')
    readonly_fields = ('group','time','offset','members_count','members_entered_count','members_left_count')
    ordering = ('-time',)
    extra = 0
    can_delete = False

    def queryset(self, request):
        qs = super(GroupMigrationInline, self).queryset(request)
        return qs.light.exclude(time__isnull=True)

class GroupAdmin(GroupAdminOriginal):
    inlines = GroupAdminOriginal.inlines + [GroupMigrationInline]

class GroupMigrationAdmin(VkontakteModelAdmin):
    list_display = ('group','time')
    list_display_links = ('time',)
#    list_filter = ('group',)

admin.site.unregister(Group)
admin.site.register(Group, GroupAdmin)
admin.site.register(GroupMigration, GroupMigrationAdmin)


from vkontakte_groups.factories import GroupFactory
from models import GroupMigration
import factory

class GroupMigrationFactory(factory.DjangoModelFactory):
    FACTORY_FOR = GroupMigration

    group = factory.SubFactory(GroupFactory)


# -*- coding: utf-8 -*-
from django.dispatch import Signal
from django.conf import settings
from annoying.decorators import signals
from vkontakte_groups.models import Group
# from models import GroupMigration
#
# group_migration_updated = Signal(providing_args=['instance'])
# group_users_updated = Signal(providing_args=['instance'])
#
# @signals(group_migration_updated, sender=GroupMigration)
# def group_users_update_m2m(sender, instance, **kwargs):
#     if 'djcelery' in settings.INSTALLED_APPS:
#         from tasks import VkontakteGroupUpdateUsersM2M
#         return VkontakteGroupUpdateUsersM2M.delay(instance.id)
#     else:
#         GroupMigration.objects.update_group_users_m2m(instance)


# -*- coding: utf-8 -*-
from django.test import TestCase
from models import RelationsHistory, NoDeltaForFirstHistory
from twitter_api.factories import UserFactory
import mock

GROUP_ID = 30221121

class TwitterRelationsHistoryMigrationTest(TestCase):

    @mock.patch('twitter_api.models.User')
    def test_relations_history(self, User, *args, **kwargs):
        user = UserFactory()
        User.remote.fetch_followers_ids_for_user.return_value = [1,2,3]
        RelationsHistory.objects.update_for_user(user)

        self.assertEqual(RelationsHistory.objects.count(), 1)
        instance1 = RelationsHistory.objects.all()[0]
        self.assertListEqual(instance1.followers_ids, [1,2,3])
        self.assertEqual(instance1.followers_count, 3)
        self.assertRaises(NoDeltaForFirstHistory, lambda: instance1.followers_entered_ids)
        self.assertRaises(NoDeltaForFirstHistory, lambda: instance1.followers_left_ids)

        User.remote.fetch_followers_ids_for_user.return_value = [1,2,3,4,5,6]
        RelationsHistory.objects.update_for_user(user)

        self.assertEqual(RelationsHistory.objects.count(), 2)
        instance2 = RelationsHistory.objects.order_by('-time')[0]
        self.assertListEqual(instance2.followers_ids, [1,2,3,4,5,6])
        self.assertListEqual(instance2.followers_entered_ids, [4,5,6])
        self.assertListEqual(instance2.followers_left_ids, [])
        self.assertEqual(instance2.followers_count, 6)
        self.assertEqual(instance2.followers_entered_count, 3)
        self.assertEqual(instance2.followers_left_count, 0)

        User.remote.fetch_followers_ids_for_user.return_value = [1,2,7]
        RelationsHistory.objects.update_for_user(user)

        self.assertEqual(RelationsHistory.objects.count(), 3)
        instance3 = RelationsHistory.objects.order_by('-time')[0]
        self.assertListEqual(instance3.followers_ids, [1,2,7])
        self.assertListEqual(instance3.followers_entered_ids, [7])
        self.assertListEqual(instance3.followers_left_ids, [3,4,5,6])
        self.assertEqual(instance3.followers_count, 3)
        self.assertEqual(instance3.followers_entered_count, 1)
        self.assertEqual(instance3.followers_left_count, 4)

        self.assertListEqual(instance3.followers_entered_ids(instance1), [7]) # TODO: implement it later
        self.assertListEqual(instance3.followers_left_ids(instance1), [3])
        self.assertEqual(instance3.followers_entered_count(instance1), 1)
        self.assertEqual(instance3.followers_left_count(instance1), 1)

#     def test_deleting_hiding_histories(self):
#
#         for i in range(1,7):
#             UserFactory.create(remote_id=i)
#
#         group = GroupFactory.create(remote_id=GROUP_ID)
#         stat1 = GroupMigration.objects.create(group=group, members_ids=[1,2,3,4,5])
#         stat1.save_final()
#         stat2 = GroupMigration.objects.create(group=group, members_ids=[1,2,3,4,6])
#         stat2.save_final()
#         stat3 = GroupMigration.objects.create(group=group, members_ids=[1,2,3,5,7])
#         stat3.save_final()
#
#         # difference between stat2 and stat1
#         self.assertEqual(stat2.members_entered_ids, [6])
#         self.assertEqual(stat2.members_left_ids, [5])
#         # difference between stat3 and stat2
#         self.assertEqual(stat3.members_entered_ids, [5,7])
#         self.assertEqual(stat3.members_left_ids, [4,6])
#
#         stat2.delete()
#         stat3 = GroupMigration.objects.get(id=stat3.id)
#
#         # difference between stat3 and stat1
#         self.assertEqual(stat3.members_entered_ids, [7])
#         self.assertEqual(stat3.members_left_ids, [4])
#
#         stat4 = GroupMigration.objects.create(group=group, members_ids=[1,2,3,4,6])
#         stat4.save_final()
#
#         # difference between stat4 and stat3
#         self.assertEqual(stat4.members_entered_ids, [4,6])
#         self.assertEqual(stat4.members_left_ids, [5,7])
#
#         stat3.hide()
#         stat4 = GroupMigration.objects.get(id=stat4.id)
#
#         # difference between stat4 and stat1
#         self.assertEqual(stat4.members_entered_ids, [6])
#         self.assertEqual(stat4.members_left_ids, [5])
#
#         stat5 = GroupMigration.objects.create(group=group, members_ids=[1,2,3,5,7])
#         stat5.save_final()
#
#         # difference between stat5 and stat4
#         self.assertEqual(stat5.members_entered_ids, [5,7])
#         self.assertEqual(stat5.members_left_ids, [4,6])
#
#         stat4.hidden = True
#         stat4.save()
#         stat5 = GroupMigration.objects.get(id=stat5.id)
#
#         # difference between stat5 and stat1
#         self.assertEqual(stat5.members_entered_ids, [7])
#         self.assertEqual(stat5.members_left_ids, [4])